import torch
import networkx as nx
import numpy as np
import os.path as osp

from cogdl.data import Dataset, Graph
from cogdl.utils import download_url, untar, Accuracy, CrossEntropyLoss


def read_geom_data(folder, dataset_name):
    graph_adjacency_list_file_path = osp.join(folder, "out1_graph_edges.txt")
    graph_node_features_and_labels_file_path = osp.join(folder, "out1_node_feature_label.txt")

    G = nx.DiGraph()
    graph_node_features_dict = {}
    graph_labels_dict = {}

    if dataset_name == "film":
        with open(graph_node_features_and_labels_file_path) as graph_node_features_and_labels_file:
            graph_node_features_and_labels_file.readline()
            for line in graph_node_features_and_labels_file:
                line = line.rstrip().split("\t")
                assert len(line) == 3
                assert int(line[0]) not in graph_node_features_dict and int(line[0]) not in graph_labels_dict
                feature_blank = np.zeros(932, dtype=np.uint8)
                feature_blank[np.array(line[1].split(","), dtype=np.uint16)] = 1
                graph_node_features_dict[int(line[0])] = feature_blank
                graph_labels_dict[int(line[0])] = int(line[2])
    else:
        with open(graph_node_features_and_labels_file_path) as graph_node_features_and_labels_file:
            graph_node_features_and_labels_file.readline()
            for line in graph_node_features_and_labels_file:
                line = line.rstrip().split("\t")
                assert len(line) == 3
                assert int(line[0]) not in graph_node_features_dict and int(line[0]) not in graph_labels_dict
                graph_node_features_dict[int(line[0])] = np.array(line[1].split(","), dtype=np.uint8)
                graph_labels_dict[int(line[0])] = int(line[2])

    with open(graph_adjacency_list_file_path) as graph_adjacency_list_file:
        graph_adjacency_list_file.readline()
        for line in graph_adjacency_list_file:
            line = line.rstrip().split("\t")
            assert len(line) == 2
            if int(line[0]) not in G:
                G.add_node(
                    int(line[0]), features=graph_node_features_dict[int(line[0])], label=graph_labels_dict[int(line[0])]
                )
            if int(line[1]) not in G:
                G.add_node(
                    int(line[1]), features=graph_node_features_dict[int(line[1])], label=graph_labels_dict[int(line[1])]
                )
            G.add_edge(int(line[0]), int(line[1]))

    adj = nx.adjacency_matrix(G, sorted(G.nodes()))
    features = np.array([features for _, features in sorted(G.nodes(data="features"), key=lambda x: x[0])])
    labels = np.array([label for _, label in sorted(G.nodes(data="label"), key=lambda x: x[0])])

    all_masks = []
    for split in range(10):
        graph_split_file_path = osp.join(folder, f"{dataset_name}_split_0.6_0.2_{split}.npz")
        with np.load(graph_split_file_path) as splits_file:
            train_mask = splits_file["train_mask"]
            val_mask = splits_file["val_mask"]
            test_mask = splits_file["test_mask"]
            train_mask = torch.BoolTensor(train_mask)
            val_mask = torch.BoolTensor(val_mask)
            test_mask = torch.BoolTensor(test_mask)
            all_masks.append({"train": train_mask, "val": val_mask, "test": test_mask})

    features = torch.FloatTensor(features)
    labels = torch.LongTensor(labels)

    coo_adj = adj.tocoo()
    row, col = torch.LongTensor(coo_adj.row), torch.LongTensor(coo_adj.col)
    edge_index = (row, col)

    data = Graph(x=features, edge_index=edge_index, y=labels, all_masks=all_masks)

    return data


class GeomDataset(Dataset):
    url = "https://cloud.tsinghua.edu.cn/d/70d8aaebf2ed493697e0/files/?p=%2F"

    def __init__(self, root, name, split=0):
        self.name = name
        self.split = split

        super(GeomDataset, self).__init__(root)

        self.data = torch.load(self.processed_paths[0])
        self.raw_dir = osp.join(self.root, self.name, "raw")
        self.processed_dir = osp.join(self.root, self.name, "processed")

        self.data.train_mask = self.data.all_masks[split]["train"]
        self.data.val_mask = self.data.all_masks[split]["val"]
        self.data.test_mask = self.data.all_masks[split]["test"]

    @property
    def raw_file_names(self):
        names = ["out1_graph_edges.txt", "out1_node_feature_label.txt"] + [
            f"{self.name}_split_0.6_0.2_{idx}.npz" for idx in range(10)
        ]
        return names

    @property
    def processed_file_names(self):
        return "data.pt"

    @property
    def num_classes(self):
        assert hasattr(self.data, "y")
        return int(torch.max(self.data.y)) + 1

    @property
    def num_nodes(self):
        assert hasattr(self.data, "y")
        return self.data.y.shape[0]

    def download(self):
        fname = "{}.zip".format(self.name.lower())
        download_url("{}{}.zip&dl=1".format(self.url, self.name.lower()), self.raw_dir, fname)
        untar(self.raw_dir, fname)

    def process(self):
        data = read_geom_data(self.raw_dir, self.name)
        torch.save(data, self.processed_paths[0])

        return data

    def get(self, idx):
        return self.data

    def __repr__(self):
        return "{}()".format(self.name)

    def __len__(self):
        return 1

    def get_evaluator(self):
        return Accuracy()

    def get_loss_fn(self):
        return CrossEntropyLoss()


class ChameleonDataset(GeomDataset):
    def __init__(self, data_path="data", split=0):
        dataset = "chameleon"
        path = osp.join(data_path, dataset)
        super(ChameleonDataset, self).__init__(path, dataset, split)


class CornellDataset(GeomDataset):
    def __init__(self, data_path="data", split=0):
        dataset = "cornell"
        path = osp.join(data_path, dataset)
        super(CornellDataset, self).__init__(path, dataset, split)


class FilmDataset(GeomDataset):
    def __init__(self, data_path="data", split=0):
        dataset = "film"
        path = osp.join(data_path, dataset)
        super(FilmDataset, self).__init__(path, dataset, split)


class SquirrelDataset(GeomDataset):
    def __init__(self, data_path="data", split=0):
        dataset = "squirrel"
        path = osp.join(data_path, dataset)
        super(SquirrelDataset, self).__init__(path, dataset, split)


class TexasDataset(GeomDataset):
    def __init__(self, data_path="data", split=0):
        dataset = "texas"
        path = osp.join(data_path, dataset)
        super(TexasDataset, self).__init__(path, dataset, split)


class WisconsinDataset(GeomDataset):
    def __init__(self, data_path="data", split=0):
        dataset = "wisconsin"
        path = osp.join(data_path, dataset)
        super(WisconsinDataset, self).__init__(path, dataset, split)



import torch.nn as nn

from actnn.conf import config
from actnn.qscheme import QScheme

from cogdl.operators.linear import linear


class QLinear(nn.Linear):
    num_layers = 0

    def __init__(self, input_features, output_features, bias=True, group=0, rp_ratio=2):
        super(QLinear, self).__init__(input_features, output_features, bias)
        if config.adaptive_conv_scheme:
            self.scheme = QScheme(self, group=group)
        else:
            self.scheme = None
        self.rp_ratio = rp_ratio

    def forward(self, input):
        if config.training:
            return linear.apply(input, self.weight, self.bias, self.scheme, self.rp_ratio)
        else:
            return super(QLinear, self).forward(input)



import math

import torch
import torch.nn as nn

from cogdl.utils import spmm, get_activation, SpMM


class QGCNLayer(nn.Module):
    """
    Simple GCN layer, similar to https://arxiv.org/abs/1609.02907
    """

    def __init__(
        self, in_features, out_features, dropout=0.0, activation=None, residual=False, norm=None, bias=True, **kwargs
    ):
        super(QGCNLayer, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.linear = nn.Linear(in_features, out_features, bias=bias)
        self.spmm = SpMM()
        # if dropout > 0:
        #     self.dropout = nn.Dropout(dropout)
        # else:
        #     self.dropout = None
        # if residual:
        #     self.residual = nn.Linear(in_features, out_features)
        # else:
        #     self.residual = None

        # if activation is not None:
        #     self.act = get_activation(activation, inplace=True)
        # else:
        #     self.act = None

        # if norm is not None:
        #     if norm == "batchnorm":
        #         self.norm = nn.BatchNorm1d(out_features)
        #     elif norm == "layernorm":
        #         self.norm = nn.LayerNorm(out_features)
        #     else:
        #         raise NotImplementedError
        # else:
        #     self.norm = None

        self.reset_parameters()

    def reset_parameters(self):
        stdv = 1.0 / math.sqrt(self.out_features)
        torch.nn.init.uniform_(self.linear.weight, -stdv, stdv)

    def forward(self, graph, x):
        support = self.linear(x)
        out = self.spmm(graph, support)
        # out = spmm(graph, support)

        # if self.norm is not None:
        #     out = self.norm(out)
        # if self.act is not None:
        #     out = self.act(out)

        # if self.residual is not None:
        #     out = out + self.residual(x)
        # if self.dropout is not None:
        #     out = self.dropout(out)
        return out



"""

"""
import numpy as np
import numpy.ctypeslib as npct
from ctypes import c_float, c_int
from os.path import dirname

array_1d_int = npct.ndpointer(dtype=np.int32, ndim=1, flags="CONTIGUOUS")

librwalk = npct.load_library("librwalk", dirname(__file__))

# print("rwalk: Loading library from: {}".format(dirname(__file__)))
# librwalk.random_walk.restype = None
# librwalk.random_walk.argtypes = [array_1d_int, array_1d_int, c_int, c_int, c_int, c_int, c_int, array_1d_int]

librwalk.random_walk.restype = None
librwalk.random_walk.argtypes = [
    array_1d_int,
    array_1d_int,
    array_1d_int,
    c_int,
    c_int,
    c_int,
    c_int,
    c_int,
    c_float,
    array_1d_int,
]


def random_walk(nodes, ptr, neighs, num_walks=1, num_steps=1, nthread=-1, seed=111413, restart_prob=0.0):
    assert ptr.flags["C_CONTIGUOUS"]
    assert neighs.flags["C_CONTIGUOUS"]
    assert ptr.dtype == np.int32
    assert neighs.dtype == np.int32
    assert nodes.dtype == np.int32
    n = nodes.size
    walks = -np.ones((n * num_walks, (num_steps + 1)), dtype=np.int32, order="C")
    assert walks.flags["C_CONTIGUOUS"]

    librwalk.random_walk(
        nodes,
        ptr,
        neighs,
        n,
        num_walks,
        num_steps,
        seed,
        nthread,
        restart_prob,
        np.reshape(walks, (walks.size,), order="C"),
    )

    return walks



import torch

from .graph_classification_dw import GraphClassificationDataWrapper
from cogdl.models.nn.patchy_san import get_single_feature


class PATCHY_SAN_DataWrapper(GraphClassificationDataWrapper):
    @staticmethod
    def add_args(parser):
        GraphClassificationDataWrapper.add_args(parser)
        parser.add_argument("--num-sample", default=30, type=int, help="Number of chosen vertexes")
        parser.add_argument("--num-neighbor", default=10, type=int, help="Number of neighbor in constructing features")
        parser.add_argument("--stride", default=1, type=int, help="Stride of chosen vertexes")

    def __init__(self, dataset, num_sample, num_neighbor, stride, *args, **kwargs):
        super(PATCHY_SAN_DataWrapper, self).__init__(dataset, *args, **kwargs)
        self.sample = num_sample
        self.dataset = dataset
        self.neighbor = num_neighbor
        self.stride = stride

    def pre_transform(self):
        dataset = self.dataset
        num_features = dataset.num_features
        num_classes = dataset.num_classes
        for i, data in enumerate(dataset):
            new_feature = get_single_feature(
                dataset[i], num_features, num_classes, self.sample, self.neighbor, self.stride
            )
            dataset[i].x = torch.from_numpy(new_feature)
        self.dataset = dataset
        super(PATCHY_SAN_DataWrapper, self).pre_transform()



import importlib

from .base_model import BaseModel


def register_model(name):
    """
    New model types can be added to cogdl with the :func:`register_model`
    function decorator.
    For example::
        @register_model('gat')
        class GAT(BaseModel):
            (...)
    Args:
        name (str): the name of the model
    """

    def register_model_cls(cls):
        print("The `register_model` API is deprecated!")
        return cls

    return register_model_cls


def try_adding_model_args(model, parser):
    if model in SUPPORTED_MODELS:
        path = ".".join(SUPPORTED_MODELS[model].split(".")[:-1])
        module = importlib.import_module(path)
        class_name = SUPPORTED_MODELS[model].split(".")[-1]
        getattr(module, class_name).add_args(parser)


def build_model(args):
    model = args.model
    if isinstance(model, list):
        model = model[0]
    if model in SUPPORTED_MODELS:
        path = ".".join(SUPPORTED_MODELS[model].split(".")[:-1])
        module = importlib.import_module(path)
    else:
        raise NotImplementedError(f"Failed to import {model} model.")
    class_name = SUPPORTED_MODELS[model].split(".")[-1]
    return getattr(module, class_name).build_model_from_args(args)


SUPPORTED_MODELS = {
    "hope": "cogdl.models.emb.hope.HOPE",
    "spectral": "cogdl.models.emb.spectral.Spectral",
    "hin2vec": "cogdl.models.emb.hin2vec.Hin2vec",
    "netmf": "cogdl.models.emb.netmf.NetMF",
    "deepwalk": "cogdl.models.emb.deepwalk.DeepWalk",
    "gatne": "cogdl.models.emb.gatne.GATNE",
    "dgk": "cogdl.models.emb.dgk.DeepGraphKernel",
    "grarep": "cogdl.models.emb.grarep.GraRep",
    "dngr": "cogdl.models.emb.dngr.DNGR",
    "prone++": "cogdl.models.emb.pronepp.ProNEPP",
    "graph2vec": "cogdl.models.emb.graph2vec.Graph2Vec",
    "metapath2vec": "cogdl.models.emb.metapath2vec.Metapath2vec",
    "node2vec": "cogdl.models.emb.node2vec.Node2vec",
    "pte": "cogdl.models.emb.pte.PTE",
    "netsmf": "cogdl.models.emb.netsmf.NetSMF",
    "line": "cogdl.models.emb.line.LINE",
    "sdne": "cogdl.models.emb.sdne.SDNE",
    "prone": "cogdl.models.emb.prone.ProNE",
    "daegc": "cogdl.models.nn.daegc.DAEGC",
    "agc": "cogdl.models.nn.agc.AGC",
    "gae": "cogdl.models.nn.gae.GAE",
    "vgae": "cogdl.models.nn.gae.VGAE",
    "dgi": "cogdl.models.nn.dgi.DGIModel",
    "mvgrl": "cogdl.models.nn.mvgrl.MVGRL",
    "patchy_san": "cogdl.models.nn.patchy_san.PatchySAN",
    "gcn": "cogdl.models.nn.gcn.GCN",
    "actgcn": "cogdl.models.nn.actgcn.ActGCN",
    "gdc_gcn": "cogdl.models.nn.gdc_gcn.GDC_GCN",
    "graphsage": "cogdl.models.nn.graphsage.Graphsage",
    "compgcn": "cogdl.models.nn.compgcn.LinkPredictCompGCN",
    "drgcn": "cogdl.models.nn.drgcn.DrGCN",
    "unet": "cogdl.models.nn.graph_unet.GraphUnet",
    "gcnmix": "cogdl.models.nn.gcnmix.GCNMix",
    "diffpool": "cogdl.models.nn.diffpool.DiffPool",
    "gcnii": "cogdl.models.nn.gcnii.GCNII",
    "sign": "cogdl.models.nn.sign.SIGN",
    "mixhop": "cogdl.models.nn.mixhop.MixHop",
    "gat": "cogdl.models.nn.gat.GAT",
    "han": "cogdl.models.nn.han.HAN",
    "ppnp": "cogdl.models.nn.ppnp.PPNP",
    "grace": "cogdl.models.nn.grace.GRACE",
    "pprgo": "cogdl.models.nn.pprgo.PPRGo",
    "gin": "cogdl.models.nn.gin.GIN",
    "grand": "cogdl.models.nn.grand.Grand",
    "gtn": "cogdl.models.nn.gtn.GTN",
    "rgcn": "cogdl.models.nn.rgcn.LinkPredictRGCN",
    "deepergcn": "cogdl.models.nn.deepergcn.DeeperGCN",
    "drgat": "cogdl.models.nn.drgat.DrGAT",
    "infograph": "cogdl.models.nn.infograph.InfoGraph",
    "dropedge_gcn": "cogdl.models.nn.dropedge_gcn.DropEdge_GCN",
    "disengcn": "cogdl.models.nn.disengcn.DisenGCN",
    "mlp": "cogdl.models.nn.mlp.MLP",
    "sgc": "cogdl.models.nn.sgc.sgc",
    "sortpool": "cogdl.models.nn.sortpool.SortPool",
    "srgcn": "cogdl.models.nn.srgcn.SRGCN",
    "gcc": "cogdl.models.nn.gcc_model.GCCModel",
    "unsup_graphsage": "cogdl.models.nn.unsup_graphsage.SAGE",
    "graphsaint": "cogdl.models.nn.graphsaint.GraphSAINT",
    "m3s": "cogdl.models.nn.m3s.M3S",
    "moe_gcn": "cogdl.models.nn.moe_gcn.MoEGCN",
    "lightgcn": "cogdl.models.nn.lightgcn.LightGCN",
    "correct_smooth_mlp": "cogdl.models.nn.correct_smooth.CorrectSmoothMLP",
    "sagn": "cogdl.models.nn.sagn.SAGN",
    "revgcn": "cogdl.models.nn.revgcn.RevGCN",
    "revgat": "cogdl.models.nn.revgcn.RevGAT",
    "revgen": "cogdl.models.nn.revgcn.RevGEN",
    "sage": "cogdl.models.nn.graphsage.SAGE",
    "autognn": "cogdl.models.nn.autognn.AutoGNN",
}



import networkx as nx
import numpy as np
import scipy.sparse as sp

from .. import BaseModel


class NetMF(BaseModel):
    r"""The NetMF model from the `"Network Embedding as Matrix Factorization: Unifying DeepWalk, LINE, PTE, and node2vec"
    <http://arxiv.org/abs/1710.02971>`_ paper.

    Args:
        hidden_size (int) : The dimension of node representation.
        window_size (int) : The actual context size which is considered in language model.
        rank (int) : The rank in approximate normalized laplacian.
        negative (int) : The number of nagative samples in negative sampling.
        is-large (bool) : When window size is large, use approximated deepwalk matrix to decompose.
    """

    @staticmethod
    def add_args(parser):
        """Add model-specific arguments to the parser."""
        # fmt: off
        parser.add_argument("--window-size", type=int, default=5)
        parser.add_argument("--rank", type=int, default=256)
        parser.add_argument("--negative", type=int, default=1)
        parser.add_argument("--is-large", action="store_true")
        parser.add_argument("--hidden-size", type=int, default=128)
        # fmt: on

    @classmethod
    def build_model_from_args(cls, args):
        return cls(args.hidden_size, args.window_size, args.rank, args.negative, args.is_large)

    def __init__(self, dimension, window_size, rank, negative, is_large=False):
        super(NetMF, self).__init__()
        self.dimension = dimension
        self.window_size = window_size
        self.rank = rank
        self.negative = negative
        self.is_large = is_large

    def forward(self, graph, return_dict=False):
        nx_g = graph.to_networkx()
        A = sp.csr_matrix(nx.adjacency_matrix(nx_g))
        if not self.is_large:
            print("Running NetMF for a small window size...")
            deepwalk_matrix = self._compute_deepwalk_matrix(A, window=self.window_size, b=self.negative)

        else:
            print("Running NetMF for a large window size...")
            vol = float(A.sum())
            evals, D_rt_invU = self._approximate_normalized_laplacian(A, rank=self.rank, which="LA")
            deepwalk_matrix = self._approximate_deepwalk_matrix(
                evals, D_rt_invU, window=self.window_size, vol=vol, b=self.negative
            )
        # factorize deepwalk matrix with SVD
        u, s, _ = sp.linalg.svds(deepwalk_matrix, self.dimension)
        embeddings = sp.diags(np.sqrt(s)).dot(u.T).T

        if return_dict:
            features_matrix = dict()
            for vid, node in enumerate(nx_g.nodes()):
                features_matrix[node] = embeddings[vid]
        else:
            features_matrix = np.zeros((graph.num_nodes, embeddings.shape[1]))
            nx_nodes = nx_g.nodes()
            features_matrix[nx_nodes] = embeddings[np.arange(graph.num_nodes)]
        return features_matrix

    def _compute_deepwalk_matrix(self, A, window, b):
        # directly compute deepwalk matrix
        n = A.shape[0]
        vol = float(A.sum())
        L, d_rt = sp.csgraph.laplacian(A, normed=True, return_diag=True)
        # X = D^{-1/2} A D^{-1/2}
        X = sp.identity(n) - L
        S = np.zeros_like(X)
        X_power = sp.identity(n)
        for i in range(window):
            print("Compute matrix %d-th power", i + 1)
            X_power = X_power.dot(X)
            S += X_power
        S *= vol / window / b
        D_rt_inv = sp.diags(d_rt ** -1)
        M = D_rt_inv.dot(D_rt_inv.dot(S).T).todense()
        M[M <= 1] = 1
        Y = np.log(M)
        return sp.csr_matrix(Y)

    def _approximate_normalized_laplacian(self, A, rank, which="LA"):
        # perform eigen-decomposition of D^{-1/2} A D^{-1/2} and keep top rank eigenpairs
        n = A.shape[0]
        L, d_rt = sp.csgraph.laplacian(A, normed=True, return_diag=True)
        # X = D^{-1/2} W D^{-1/2}
        X = sp.identity(n) - L
        print("Eigen decomposition...")
        evals, evecs = sp.linalg.eigsh(X, rank, which=which)
        print("Maximum eigenvalue %f, minimum eigenvalue %f", np.max(evals), np.min(evals))
        print("Computing D^{-1/2}U..")
        D_rt_inv = sp.diags(d_rt ** -1)
        D_rt_invU = D_rt_inv.dot(evecs)
        return evals, D_rt_invU

    def _deepwalk_filter(self, evals, window):
        for i in range(len(evals)):
            x = evals[i]
            evals[i] = 1.0 if x >= 1 else x * (1 - x ** window) / (1 - x) / window
        evals = np.maximum(evals, 0)
        print(
            "After filtering, max eigenvalue=%f, min eigenvalue=%f",
            np.max(evals),
            np.min(evals),
        )
        return evals

    def _approximate_deepwalk_matrix(self, evals, D_rt_invU, window, vol, b):
        # approximate deepwalk matrix
        evals = self._deepwalk_filter(evals, window=window)
        X = sp.diags(np.sqrt(evals)).dot(D_rt_invU.T).T
        M = X.dot(X.T) * vol / b
        M[M <= 1] = 1
        Y = np.log(M)
        print("Computed DeepWalk matrix with %d non-zero elements", np.count_nonzero(Y))
        return sp.csr_matrix(Y)



from .base_layer import BaseLayer
from .gcn_layer import GCNLayer
from .sage_layer import MeanAggregator, SumAggregator, SAGELayer
from .gat_layer import GATLayer
from .gin_layer import GINLayer
from .gine_layer import GINELayer
from .se_layer import SELayer
from .deepergcn_layer import GENConv, ResGNNLayer
from .disengcn_layer import DisenGCNLayer
from .gcnii_layer import GCNIILayer
from .mlp_layer import MLP
from .saint_layer import SAINTLayer
from .han_layer import HANLayer
from .pprgo_layer import PPRGoLayer
from .rgcn_layer import RGCNLayer
from .sgc_layer import SGCLayer
from .mixhop_layer import MixHopLayer
from .reversible_layer import RevGNNLayer
from .set2set import Set2Set


__all__ = [
    "BaseLayer",
    "GCNLayer",
    "MeanAggregator",
    "SumAggregator",
    "SAGELayer",
    "GATLayer",
    "GINLayer",
    "GINELayer",
    "SELayer",
    "GENConv",
    "ResGNNLayer",
    "DisenGCNLayer",
    "GCNIILayer",
    "SAINTLayer",
    "HANLayer",
    "PPRGoLayer",
    "RGCNLayer",
    "SGCLayer",
    "MixHopLayer",
    "MLP",
    "RevGNNLayer",
    "Set2Set",
]



from cogdl.layers import SGCLayer

from .. import BaseModel


class sgc(BaseModel):
    @staticmethod
    def add_args(parser):
        parser.add_argument("--num-features", type=int)
        parser.add_argument("--num-classes", type=int)

    @classmethod
    def build_model_from_args(cls, args):
        return cls(in_feats=args.num_features, out_feats=args.num_classes)

    def __init__(self, in_feats, out_feats):
        super(sgc, self).__init__()
        self.nn = SGCLayer(in_feats, out_feats)
        self.cache = dict()

    def forward(self, graph):
        graph.sym_norm()

        x = self.nn(graph, graph.x)
        return x

    def predict(self, data):
        return self.forward(data)



from typing import List, Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F
from cogdl.data import Graph
from cogdl.layers import GCNLayer
from cogdl.utils import dropout_adj, get_activation
from torch_sparse import spspmm

from .. import BaseModel


class Pool(nn.Module):
    def __init__(
        self, in_feats: int, pooling_rate: float, aug_adj: bool = False, dropout: float = 0.5, activation: str = "tanh"
    ):
        super(Pool, self).__init__()
        self.aug_adj = aug_adj
        self.pooling_rate = pooling_rate
        self.act = get_activation(activation)
        self.proj = nn.Linear(in_feats, 1)
        self.dropout = nn.Dropout(dropout) if dropout > 0 else nn.Identity

    def forward(self, graph: Graph, x: torch.Tensor) -> Tuple[Graph, torch.Tensor]:
        x = self.dropout(x)
        h = self.proj(x).squeeze()
        scores = self.act(h)
        return self.top_k(graph, x, scores)

    def top_k(self, graph, x: torch.Tensor, scores: torch.Tensor) -> Tuple[Graph, torch.Tensor]:
        org_n_nodes = x.shape[0]
        num = int(self.pooling_rate * x.shape[0])
        values, indices = torch.topk(scores, max(2, num))

        if self.aug_adj:
            edge_attr = torch.ones(graph.edge_index[0].shape[0])
            edge_index = torch.stack(graph.edge_index).cpu()
            edge_index, _ = spspmm(edge_index, edge_attr, edge_index, edge_attr, org_n_nodes, org_n_nodes, org_n_nodes)
            edge_index = edge_index.to(x.device)
            batch = Graph(x=x, edge_index=edge_index)
        else:
            batch = graph
            batch.x = x
        new_batch = batch.subgraph(indices)

        new_batch.row_norm()
        return new_batch, indices


class UnPool(nn.Module):
    def __init__(self):
        super(UnPool, self).__init__()

    def forward(self, num_nodes: int, h: torch.Tensor, indices: torch.Tensor) -> torch.Tensor:
        new_h = torch.zeros(num_nodes, h.shape[1]).to(h.device)
        new_h[indices] = h
        return new_h


class GraphUnetLayer(nn.Module):
    def __init__(
        self,
        hidden_size: int,
        pooling_layer: int,
        pooling_rates: List[float],
        activation: str = "elu",
        dropout: float = 0.5,
        aug_adj: bool = False,
    ):
        super(GraphUnetLayer, self).__init__()
        self.dropout = dropout
        self.activation = activation
        self.pooling_layer = pooling_layer
        self.gcn = GCNLayer(hidden_size, hidden_size)
        self.act = get_activation(activation)

        self.down_gnns = nn.ModuleList([GCNLayer(hidden_size, hidden_size) for _ in range(pooling_layer)])
        self.up_gnns = nn.ModuleList([GCNLayer(hidden_size, hidden_size) for _ in range(pooling_layer)])
        self.poolings = nn.ModuleList(
            [Pool(hidden_size, pooling_rates[i], aug_adj, dropout) for i in range(pooling_layer)]
        )
        self.unpoolings = nn.ModuleList([UnPool() for _ in range(pooling_layer)])

    def forward(self, graph: Graph, x: torch.Tensor) -> List[torch.Tensor]:
        adjs = []
        indices = []
        down_hidden = []
        num_nodes = []
        h_list = []

        h_init = x
        h = x

        g = graph
        for i in range(self.pooling_layer):
            h = F.dropout(h, p=self.dropout, training=self.training)
            h = self.down_gnns[i](g, h)
            h = self.act(h)

            adjs.append(g)
            down_hidden.append(h)
            num_nodes.append(h.shape[0])

            g, index = self.poolings[i](g, h)
            h = g.x
            indices.append(index)

        h = F.dropout(h, p=self.dropout, training=self.training)
        h = self.gcn(g, h)
        h = self.act(h)

        for _i in range(self.pooling_layer):
            i = self.pooling_layer - _i - 1
            g = adjs[i]
            index = indices[i]
            h = self.unpoolings[i](num_nodes[i], h, index)

            h = F.dropout(h, p=self.dropout, training=self.training)
            h = self.up_gnns[i](g, h)
            h = self.act(h)

            h = h + down_hidden[i]
            h_list.append(h)
        h = h.add(h_init)
        h_list.append(h)
        return h_list


class GraphUnet(BaseModel):
    @staticmethod
    def add_args(parser):
        # fmt: off
        parser.add_argument("--hidden-size", type=int, default=128)
        parser.add_argument("--n-dropout", type=float, default=0.8)

        parser.add_argument("--adj-dropout", type=float, default=0.0)
        parser.add_argument("--n-pool", type=int, default=4)
        parser.add_argument("--pool-rate", nargs="+", default=[0.7, 0.5, 0.5, 0.4])
        parser.add_argument("--activation", type=str, default="relu")

        parser.add_argument("--improved", action="store_true")
        parser.add_argument("--aug-adj", action="store_true")
        # fmt: on

    @classmethod
    def build_model_from_args(cls, args):
        return cls(
            in_feats=args.num_features,
            hidden_size=args.hidden_size,
            out_feats=args.num_classes,
            pooling_layer=args.n_pool,
            pooling_rates=args.pool_rate,
            n_dropout=args.n_dropout,
            adj_dropout=args.adj_dropout,
            activation=args.activation,
            improved=args.improved,
            aug_adj=args.aug_adj,
        )

    def __init__(
        self,
        in_feats: int,
        hidden_size: int,
        out_feats: int,
        pooling_layer: int,
        pooling_rates: List[float],
        n_dropout: float = 0.5,
        adj_dropout: float = 0.3,
        activation: str = "elu",
        improved: bool = False,
        aug_adj: bool = False,
    ):
        super(GraphUnet, self).__init__()
        self.improved = improved
        self.n_dropout = n_dropout
        self.adj_dropout = adj_dropout

        self.act = get_activation(activation)
        assert pooling_layer <= len(pooling_rates)
        pooling_rates = pooling_rates[:pooling_layer]
        pooling_rates = [float(x) for x in pooling_rates]
        self.unet = GraphUnetLayer(hidden_size, pooling_layer, pooling_rates, activation, n_dropout, aug_adj)

        self.in_gcn = GCNLayer(in_feats, hidden_size)
        self.out_gcn = GCNLayer(hidden_size, out_feats)

    def forward(self, graph: Graph) -> torch.Tensor:
        x = graph.x
        if self.improved and not hasattr(graph, "unet_improved"):
            row, col = graph.edge_index
            row = torch.cat([row, torch.arange(0, x.shape[0], device=x.device)], dim=0)
            col = torch.cat([col, torch.arange(0, x.shape[0], device=x.device)], dim=0)
            graph.edge_index = (row, col)
            graph["unet_improved"] = True
        graph.row_norm()

        with graph.local_graph():
            if self.training and self.adj_dropout > 0:
                graph.edge_index, graph.edge_weight = dropout_adj(graph.edge_index, graph.edge_weight, self.adj_dropout)

            x = F.dropout(x, p=self.n_dropout, training=self.training)
            h = self.in_gcn(graph, x)
            h = self.act(h)
            h_list = self.unet(graph, h)

            h = h_list[-1]
            h = F.dropout(h, p=self.n_dropout, training=self.training)
            return self.out_gcn(graph, h)
